// interpol.cpp
// Implementacin de la clase Interpol

#include <error.h>
#include <matutil.h>
#include <interpol.h>

// Constructor 
Interpol::Interpol() 
{
    // Grado mnimo posible
    pGrado = 0;

    // Aade interruptores (abiertos por defecto)
    // an no hemos establecido los datos
    pCC.nuevoInterruptor("datos"); 
    // an no hemos establecido el polinomio
    pCC.nuevoInterruptor("polinomio"); 
    // an no hemos interpolado
    pCC.nuevoInterruptor("interpol"); 

}

// destructor
Interpol::~Interpol() 
{
}
int Interpol::grado() const{
    return pGrado;
}
// devuelve los datos
const DVector& Interpol::x() const{
    return pX;
}
const DVector& Interpol::fx() const{
    return pY;
}
// devuelve el vector de coeficientes actualmente interpolado
const DVector& Interpol::ci() const{
    return pC;
}

// Establece datos
void Interpol::eDatos(const CDatos& datos){
    if (pCC.interruptorAbierto("datos"))
	pCC.cierraInterruptor("datos");
    pX = datos.x();
    pY = datos.y();

    // nuevos datos, hay que interpolar de nuevo
    pCC.abreInterruptor("interpol");
}

// Establece grado del polinomio
// (interpolamos un polinomio a los puntos en el intervalo [i,j]
void Interpol::ePolinomio(int i, int j){
    if (i > j ) 
	error("error en Interpol::ePolinomio, i < j",CERANGO);

    // Contrae el vector hasta vaciarlo
    if (pCC.interruptorCerrado("polinomio")){
	while (!pC.vacio())
	    pC.contrae();
    }
    else{
	pCC.cierraInterruptor("polinomio");
    }
    // establece rango e intervalo
    pGrado = j - i;
    pInf = i;
    pSup = j;
    
    for (int k = i; k<= j; k++){
	// f(x) de los datos
	double y = pY[k];   // intervalo vara desde [x0,x1...xn]
	// expande el vector de coeficientes
	pC.expande(y);
    }	

    // nuevo polinomio hay que interpolar de nuevo
    pCC.abreInterruptor("interpol");

}

// Acciones
// valor del polinomio actualmente interpolado en x
double Interpol::p(double x){

    if (pCC.interruptorAbierto("datos"))
	error("Interpol::p -- Datos no establecidos",
	      CENOINI);

    if (pCC.interruptorAbierto("polinomio"))
	error("Interpol::p -- Polinomio no establecido",
	      CENOINI);

    if (pCC.interruptorAbierto("interpol")){
	calculaCoeficientes(pX,pC,pInf,pSup);
	pCC.cierraInterruptor("interpol");
    }

    double px = pC[0];  // init p = c00

    for (int i = 1; i < pC.dimension(); i++){
	double prod = 1;
	for (int j = 0; j <= i-1; j++){
	    // interpolamos en el rango [pInf,pSup]
	    double xj = pX[j+pInf]; 

	    // c00 + c01(x-x0) + c02(x-x0)(x-x1)++++
	    prod *= (x - xj);
	}
	double ci = pC[i]; // coeficientes c00, c01, c02...
	px += ci*prod;
    }
    return px;
		    
}
 
// Calcula coeficientes del polinomio interpolado
// (mediante diferencias divididas)
void Interpol::calculaCoeficientes(const DVector& x, DVector& c,
				    int ji, int js)
{
    if (js -ji +1 != c.dimension())
	error("Interpol::calculaCoeficientes dim x != dim c",
	      CEMALDIM);

    // Los coeficientes que hemos almacenado por ahora
    // son c0 = f[x0], c1 = f[x1], c2 = f[x2]... cn = f[xn]
    // la siguiente columna de diferencias divididas se calcula usando
    // estos coeficientes. Si empezamos por abajo:
    // f[xn-1,xn] necesita f[xn] y f[xn-1]. Una vez calculado el trmino 
    // f[xn-1,xn] no necesitamos f[xn] y podemos utilizar el espacio:
    // cn = f[xn-1,xn]. Anlogamente, una va calculado el trmino
    // f[xn-2,xn-1] no necesitamos f[xn-1] y podemos utilizar el espacio:
    // cn-1 = f[xn-2,xn-1]. Subimos a lo largo de la columna hasta almacenar
    // f[x0,x1] en c1. El trmino c0 = f[x0] que es el primer trmino
    // en la frmula de Newton de las diferencias divididas.
    // Repitiendo el mismo proceso columna a columna, nos van quedando
    // c1 = f[x0,x1], c2 = f[x0,x1,x2], etc., hasta completar el vector
    // de coeficientes.
    
    int rango = js -ji ;

    for (int j = 1; j <= rango; j++){
	for (int i = rango; i >= j; i--){
	    double xi = x[i+ji];
	    double ximj = x[i+ji-j];
	    c[i] = (c[i] - c[i-1])/(xi - ximj);
	}
    }
}
